<?php

/**
 * @file The transaction API.
 */

/**
 * Base class with common functionality.
 */
class iDEALBase {
  function __construct(array $properties) {
    foreach ($properties as $property => $value) {
      if (property_exists($this, $property)) {
        $this->$property = $value;
      }
    }
  }
}

/**
 * An iDEAL transaction. Contains all transaction-specific data.
 */
class iDEALTransaction extends iDEALBase {

  /**
   * The internal ID of this transaction.
   *
   * @var integer
   */
  public $itid = 0;

  /**
   * The merchant that is used to make this transaction.
   *
   * @var iDEALMerchant
   */
  public $merchant = NULL;

  /**
   * The transaction amount.
   *
   * @var integer
   */
  public $amount = 0;

  /**
   * The transaction description.
   *
   * @var string
   */
  public $description = NULL;

  /**
   * The purchase ID.
   *
   * @var string
   */
  public $purchase_id = NULL;

  /**
   * The transaction ID.
   *
   * @var integer
   */
  public $transaction_id = NULL;

  /**
   * The current status.
   *
   * @var string
   */
  public $status = IDEAL_STATUS_OPEN;

  /**
   * The Unix datetime the transaction request was created.
   *
   * @var integer
   */
  public $created = 0;

  /**
   * The Unix datetime the transaction was last updated to the DB.
   *
   * @var integer
   */
  public $updated = 0;

  /**
   * The entrance code, which should be NULL if $status != IDEAL_STATUS_OPEN.
   *
   * @var string
   */
  public $entrance_code = NULL;

  /**
   * The destination after a successful transaction, which should be NULL if
   * $status != IDEAL_STATUS_OPEN.
   *
   * @var string
   */
  public $redirect_success = NULL;

  /**
   * The destination after a failed transaction, which should be NULL if
   * $status != IDEAL_STATUS_OPEN.
   *
   * @var string
   */
  public $redirect_error = NULL;

  /**
   * The payer's real name.
   *
   * @var string
   */
  public $name = NULL;

  /**
   * The city the payer lives in.
   *
   * @var string
   */
  public $city = NULL;

  /**
   * The payer's bank account number.
   *
   * @var integer
   */
  public $account_number = NULL;

  /**
   * The expiration period, e.g. the number of seconds an open transaction
   * stays valid.
   *
   * @var integer
   */
  public $expiration_period = NULL;

  /**
   * The URL as which the user can approve the transaction.
   *
   * @var string
   */
  public $issuer_authentication_url = NULL;

  /**
   * Fill the properties with passed on values.
   *
   * @param $properties array
   *   An associative array. Keys are property names and values are property
   *   values.
   */
  function __construct(array $properties = array()) {
    parent::__construct($properties);
    if (!$this->merchant) {
      $this->merchant = ideal_merchant_load(variable_get('ideal_default_merchant', 1));
    }
    if (!$this->expiration_period) {
      $this->expiration_period = variable_get('ideal_default_expiration_period', 900);
    }
  }

  /**
   * Return the entrance code for a transaction request.
   *
   * @return string
   */
  function entranceCode() {
    if (!$this->entrance_code) {
      $characters = 'abcdefghijklmnoprstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';
      for ($i = 0; $i < 40; $i++) {
        $this->entrance_code .= $characters[mt_rand(0, strlen($characters) - 1)];
      }
    }

    return $this->entrance_code;
  }

  /**
   * Start the actual transaction.
   *
   * @param $issuer_id string
   *   The issuing bank's ID, with leading zeroes until the string is 4
   *   characters long.
   *
   * @return boolean
   *   FALSE if the transaction could not be executed. There is no return value
   * on success, because the user will be redirected.
   */
  function pay($issuer_id) {
    $this->entranceCode();
    if ($response = $this->merchant->acquirer->transactionRequest($this, $issuer_id)) {
      if ($response) {
        drupal_goto($response);
      }
      else {
        drupal_goto($redirect_error);
      }
    }
    return FALSE;
  }

  /**
   * Receive a consumer after he came back from his issuer's website.
   */
  function receiveConsumer() {
    if ($this->merchant->acquirer->statusRequest($this) && $this->status == IDEAL_STATUS_SUCCESS) {
      drupal_goto($this->redirect_success);
    }
    else {
      drupal_goto($this->redirect_error);
    }
  }
}

/**
 * Merchant configuration.
 */
class iDEALMerchant extends iDEALBase {

  /**
   * Whether this merchant is locked and cannot be edited.
   *
   * @var boolean
   */
  public $locked = FALSE;

  /**
   * The internal ID of this merchant.
   *
   * @var integer
   */
  public $imid = 0;

  /**
   * The human-readable title.
   *
   * @var string
   */
  public $title = NULL;

  /**
   * The merchant ID as provided by the bank.
   *
   * @var string
   */
  public $merchant_id = NULL;

  /**
   * The sub ID.
   *
   * @var integer
   */
  public $sub_id = 0;

  /**
   * The path to the private key file.
   *
   * @var string
   */
  public $private_key_path = NULL;

  /**
   * The private key password.
   *
   * @var string
   */
  public $private_key_password = NULL;

  /**
   * The path to the private certificate file.
   *
   * @var integer
   */
  public $private_certificate_path = NULL;

  /**
   * The acquirer this merchant uses.
   *
   * @var iDEALAcquirer
   */
  public $acquirer = NULL;

  /**
   * The IAID of $this->acquirer. Only used for storage purposes.
   *
   * @var integer
   */
  public $iaid = 0;

  function __construct(array $properties = array()) {
    parent::__construct($properties);
    if (!$this->acquirer && $this->iaid) {
      $this->acquirer = ideal_acquirer_load($this->iaid);
    }
  }
}

/**
 * Acquirer configuration.
 */
class iDEALAcquirer extends iDEALBase {

  /**
   * Whether this acquirer is locked and cannot be edited.
   *
   * @var boolean
   */
  public $locked = FALSE;

  /**
   * The acquirer's internal ID.
   *
   * @var integer
   */
  public $iaid = 0;

  /**
   * The acquirer's human-readable title.
   *
   * @var string
   */
  public $title = NULL;

  /**
   * The path to the acquirer's public certificate file.
   *
   * @var integer
   */
  public $public_certificate_path = NULL;

  /**
   * The URL of the live/production server.
   *
   * @var string
   */
  public $url_live = NULL;

  /**
   * The URL of the testing/staging server.
   *
   * @var string
   */
  public $url_test = NULL;

  /**
   * Prepare a description for a request.
   *
   * @param $description string
   *
   * @return string
   */
  protected function description($description) {
    // Convert diacritics to HTML entities and replace illegal characters with
    // underscores.
    return trim(preg_replace('#[^a-z0-9=[:space:]%*+,-./&@"\':;?()$]#i', '_', htmlentities($description, ENT_NOQUOTES, 'UTF-8')));
  }

  /**
   * Construct the XML for a request.
   *
   * @param $fields array
   *   A nested array representing the XML tree. Keys are tags, values are
   *   tags' contents.
   *
   * @return string
   *   XML code.
   */
  private function xml(array $fields) {
    $xml = '';
    foreach ($fields as $name => $value) {
      $xml .= "<$name>";
      if (is_array($value)) {
        $value = $this->xml($value);
      }
      $xml .= "$value</$name>";
    }

    return $xml;
  }

  /**
   * Create a tokenCode (signature) for a request.
   *
   * @param $values array
   *   The values to create a tokenCode from.
   * @param $private_key_path string
   *   The path to the file that contains the merchant's private key.
   * @param $private_key_password string
   *   The password with which to decrypt the merchant's private key.
   *
   * @return mixed
   *   The tokenCode as a string or FALSE in case of failure.
   */
  function tokenCode(array $values, $private_key_path, $private_key_password = NULL) {
    $private_key = openssl_pkey_get_private('file://' . $private_key_path, $private_key_password);
    if ($private_key) {
      $data = preg_replace('#\s#', '', implode($values));
      $token_code = NULL;
      openssl_sign($data, $token_code, $private_key);
      openssl_free_key($private_key);
      return base64_encode($token_code);
    }
    watchdog('iDEAL', 'The private key could not be extracted from private key file file %file.', array(
      '%file' => $public_certificate_path,
    ), WATCHDOG_ERROR);
    return FALSE;
  }

  /**
   * Verify a received tokenCode (signature) for a response.
   *
   * @param $values array
   *   The values the tokenCode was created from.
   * @param $token_code string
   *   The tokenCode to check.
   * @param $public_certificate_path string
   *   The path to the file that contains the acquirer's public certificate.
   *
   * @return boolean
   *   TRUE if the tokenCode is valid. FALSE if the tokenCode is invalid or
   *   verification could not be done.
   */
  function verifyTokenCode($values, $token_code, $public_certificate_path) {
    $public_key = openssl_pkey_get_public('file://' . $public_certificate_path);
    if ($public_key) {
      $data = preg_replace('#\s#', '', implode($values));
      $return = (bool) openssl_verify($data, base64_decode($token_code), $public_key);
      openssl_free_key($public_key);
      return $return;
    }
    watchdog('iDEAL', 'The public certificate could not be extracted from public certificate file %file.', array(
      '%file' => $public_certificate_path,
    ), WATCHDOG_ERROR);
    return FALSE;
  }

  /**
   * Create the certificate's fingerprint for a request's token field.
   *
   * @param $private_certificate_path string
   *   The path to the file that contains the merchant's private certificate.
   *
   * @return string
   */
  function token($private_certificate_path) {
    $certificate = NULL;
    if (openssl_x509_export(openssl_x509_read(file_get_contents($private_certificate_path)), $certificate)) {
      $certificate = str_replace(array('-----BEGIN CERTIFICATE-----', '-----END CERTIFICATE-----'), '', $certificate);
      $certificate = base64_decode($certificate);
      $fingerprint = sha1($certificate);
      return strtoupper($fingerprint);
    }
    watchdog('iDEAL', 'The private certificate could not be extracted from private certificate file file %file.', array(
      '%file' => $public_certificate_path,
    ), WATCHDOG_ERROR);
    return FALSE;
  }

  /**
   * Construct the XML for a request.
   *
   * @param $merchant iDEALMerchant
   *   The merchant to use for this request.
   * @param $request_type string
   *   Either 'Directory', 'AcquirerTrx' (transaction), or 'AcquirerStatus'.
   * @param $token_code_values array
   *   An array of field values required to generate the token code.
   * @param $fields array
   *   A nested array representing the XML tree. Keys are tags, values are
   *   tags' contents.
   *
   * @return mixed
   *   Te response data or FALSE if the request failed.
   */
  protected function request(iDEALMerchant $merchant, $request_type, array $token_code_values, array $fields = array()) {
    // Prefix the merchant ID with zeroes until it is 9 characters long.
    $merchant_id = str_repeat('0', 9 - strlen($merchant->merchant_id)) . $merchant->merchant_id;
    $datetimestamp = gmdate('o-m-d\TH:i:s.000\Z', time());
    array_unshift($token_code_values, $datetimestamp);
    $defaults = array(
      'createDateTimeStamp' => $datetimestamp,
      'Merchant' => array(
        'merchantID' => $merchant_id,
        'subID' => $merchant->sub_id,
        'authentication' => 'SHA1_RSA',
        'token' => $this->token($merchant->private_certificate_path),
        'tokenCode' => $this->tokenCode($token_code_values, $merchant->private_key_path, $merchant->private_key_password),
      ),
    );
    $fields = array_merge_recursive($fields, $defaults);
    $xml = '<?xml version="1.0" encoding="UTF-8"?>';
    $xml .= '<' . $request_type . 'Req xmlns="http://www.idealdesk.com/Message" version="1.1.0">';
    $xml .= $this->xml($fields);
    $xml .= '</' . $request_type . 'Req>';

    $url = variable_get('ideal_test', TRUE) ? $this->url_test : $this->url_live;
    $response = drupal_http_request($url, array('content-type' => 'text/xml; charset="UTF-8"'), 'POST', $xml);

    // The HTTP request went wrong.
    if (isset($response->error)) {
      watchdog('iDEAL', '@error_message (@error_code)', array(
        '@error_message' => $response->error,
        '@error_code' => $response->code,
      ), WATCHDOG_ERROR);
      return FALSE;
    }
    $dom = simplexml_load_string($response->data);
    // The iDEAL server encountered an error.
    if (isset($dom->Error)) {
      watchdog('iDEAL', '@error_message (@error_code)', array(
        '@error_message' => $dom->Error->errorDetail,
        '@error_code' => $dom->Error->errorCode,
      ), WATCHDOG_ERROR);
      return FALSE;
    }
    // Everything went allright.
    return $response->data;
  }

  /**
   * Execute a directory request.
   *
   * @param $merchant iDEALMerchant
   *   The merchant to use for this request.
   *
   * @return array
   *  Should be of the form array(
   *    'Short' => array(
   *      'issuerID' => 'issuerName',
   *    ),
   *    'Long' => array(
   *      'issuerID' => 'issuerName',
   *    ),
   *  );
   */
  protected function directoryRequest(iDEALMerchant $merchant) {
    $token_code_values = array(
      $merchant->merchant_id,
      $merchant->sub_id,
    );
    if ($xml = $this->request($merchant, 'Directory', $token_code_values, array())) {
      $dom = simplexml_load_string($xml);
      $issuers = array('Short' => array(), 'Long' => array());
      foreach ($dom->Directory->Issuer as $issuer) {
        $issuers[(string) $issuer->issuerList][(string) $issuer->issuerID] = (string) $issuer->issuerName;
      }
      return $issuers;
    }
    return FALSE;
  }

  /**
   * Return a list of issuing banks.
   *
   * @param $merchant iDEALMerchant
   *   The merchant to use for this request.
   */
  function issuers(iDEALMerchant $merchant) {
    $cache = cache_get('ideal_issuers');
    if ($cache) {
      return $cache->data;
    }
    else {
      $issuers = $this->directoryRequest($merchant);
      // Cache issuers for 24h in case of succes. If the request fails, cache
      // the result for ten minutes to prevent errors from repeatedly popping
      // up.
      $lifetime = 86400;
      if (!$issuers || empty($issuers['Short'])) {
        $lifetime = 600;
        watchdog('iDEAL', 'iDEAL could not retrieve the list of issuing banks and will try again in ten minutes.', NULL, WATCHDOG_ERROR);
      }
      cache_set('ideal_issuers', $issuers, 'cache', time() + $lifetime);
      return $issuers;
    }
  }

  /**
   * Execute a transaction request.
   *
   * @param $transaction iDEALTransaction
   *   The transaction to execute a request for.
   * @param $issuer_id string
   *   The issuing bank's ID, with leading zeroes until the string is 4
   *   characters long.
   *
   * @return mixed
   *   The issuer's authentication URL or FALSE in case of an error.
   */
  function transactionRequest(iDEALTransaction $transaction, $issuer_id) {
    global $language;

    // Check for required data. Assume merchant configuration is in order.
    foreach (array('amount', 'description', 'purchase_id') as $property) {
      if (!$transaction->$property) {
        watchdog('iDEAL', 'Missing required transaction property %property.', array(
          '%property' => '$transaction->' . $property,
        ), WATCHDOG_ERROR);
        return FALSE;
      }
    }

    $description = $this->description($transaction->description);
    $token_code_fields = array(
      $issuer_id,
      $transaction->merchant->merchant_id,
      $transaction->merchant->sub_id,
      ideal_merchant_return_url(),
      $transaction->purchase_id,
      $transaction->amount,
      'EUR',
      $language->language,
      $description,
      $transaction->entranceCode(),
    );
    $fields = array(
      'Issuer' => array(
        'issuerID' => $issuer_id,
      ),
      'Merchant' => array(
        'merchantReturnURL' => ideal_merchant_return_url(),
      ),
      'Transaction' => array(
        'purchaseID' => $transaction->purchase_id,
        'amount' => $transaction->amount,
        'currency' => 'EUR',
        'expirationPeriod' => ideal_secondstoiso8601($transaction->expiration_period),
        'language' => $language->language,
        'description' => $description,
        'entranceCode' => $transaction->entranceCode(),
      ),
    );

    if ($xml = $this->request($transaction->merchant, 'AcquirerTrx', $token_code_fields, $fields)) {
      $dom = simplexml_load_string($xml);
      $transaction->transaction_id = (string) $dom->Transaction->transactionID;
      $transaction->issuer_authentication_url = (string) $dom->Issuer->issuerAuthenticationURL;
      $transaction->created = time();
      ideal_transaction_save($transaction);
      return (string) $dom->Issuer->issuerAuthenticationURL;
    }
    return FALSE;
  }

  /**
   * Execute a status request.
   *
   * @param $transaction iDEALTransaction
   *   The transaction to check the status for.
   *
   * @return boolean
   *   Whether the transaction has been completed.
   */
  function statusRequest(iDEALTransaction $transaction) {
    $token_code_fields = array(
      $transaction->merchant->merchant_id,
      $transaction->merchant->sub_id,
      $transaction->transaction_id,
    );
    $fields = array(
      'transactionID' => $transaction->transaction_id,
    );

    if ($xml = $this->request($transaction->merchant, 'AcquirerStatus', $token_code_fields, $fields)) {
      $dom = simplexml_load_string($xml);

      // Verify the response's authenticity.
      $token_code_values = array(
        (string) $dom->createDateTimeStamp,
        (string) $dom->Transaction->transactionID,
        (string) $dom->Transaction->status,
      );
      $status = (string) $dom->Transaction->status;
      // The account number is only added in case of success.
      if ($status == IDEAL_STATUS_SUCCESS) {
        $token_code_values[] = (string) $dom->Transaction->consumerAccountNumber;
      }
      if ($this->verifyTokenCode($token_code_values, (string) $dom->Signature->signatureValue, $transaction->merchant->acquirer->public_certificate_path)) {
        // The response is authentic. Update $transaction.
        $transaction->status = $status;
        if ($status == IDEAL_STATUS_SUCCESS) {
          $transaction->name = (string) $dom->Transaction->consumerName;
          $transaction->city = (string) $dom->Transaction->consumerCity;
          $transaction->account_number = (string) $dom->Transaction->consumerAccountNumber;
        }
        ideal_transaction_save($transaction);
        if ($status != IDEAL_STATUS_OPEN) {
          return TRUE;
        }
      }
    }
    return FALSE;
  }
}

/**
 * Display a transaction form.
 *
 * @param $transaction iDEALTransaction
 * @param $description string
 *   A description to show to the user about the transaction that he is about
 *   to make.
 * @param $minimum_amount mixed
 *   An integer that represents the minimum amount of cents a user needs to
 *   enter or FALSE to use a fixed transaction amount.
 *
 * @return array
 *   A Drupal form.
 */
function ideal_form_transaction(array $form_state, iDEALTransaction $transaction, $description, $minimum_amount = FALSE) {
  $form = array();
  if (!($issuers = $transaction->merchant->acquirer->issuers($transaction->merchant))) {
    drupal_set_message(t('The list of available banks could not be retrieved.'), 'error');
    $form['#access'] = FALSE;
  }

  // Explicitely set form handlers to make sure they will be called, even if
  // the form is embedded in another form.
  $form['#submit'] = array('ideal_form_transaction_submit');
  $form['#validate'] = array('ideal_form_transaction_validate');
  $form['#theme'] = 'ideal_form_transaction';

  $form['transaction'] = array(
    '#type' => 'value',
    '#value' => $transaction,
  );
  $form['description'] = array(
    '#value' => $description,
  );
  if ($minimum_amount) {
    $form['amount'] = array(
      '#type' => 'ideal_transaction_amount',
      '#title' => t('Amount'),
      '#amount' => $transaction->amount,
      '#minimum_amount' => $minimum_amount,
      '#access' => $minimum_amount !== FALSE,
    );
  }
  else {
    $form['amount'] = array(
      '#type' => 'value',
      '#value' => $transaction->amount,
    );
    $form['amount_human_readable'] = array(
      '#type' => 'item',
      '#title' => t('Amount'),
      '#value' => '€' . ideal_cents_to_human_readable($transaction->amount),
    );
  }
  $options = array(
    'ideal_choose' => t('Choose your bank') . '...',
  );
  if (isset($issuers['Short'])) {
    foreach ($issuers['Short'] as $issuer_id => $issuer_name) {
      $options[$issuer_id] = $issuer_name;
    }
  }
  if (!empty($issuers['Long'])) {
    $options['ideal_choose_other'] = '---' . t('Other banks') . '---';
    foreach ($issuers['Long'] as $issuer_id => $issuer_name) {
      $options[$issuer_id] = $issuer_name;
    }
  }
  $form['issuer'] = array(
    '#type' => 'select',
    '#title' => t('Choose your bank'),
    '#options' => $options,
    '#required' => TRUE,
  );
  $form['return'] = array(
    '#value' =>  t('You will automatically return to @site_name after completing the transaction.', array('@site_name' => variable_get('site_name', 'Drupal'))),
  );
  if (variable_get('ideal_test', TRUE)) {
    $form['test'] = array(
      '#value' =>  t('This is a test transaction. No actual money will be transferred.'),
    );
  }
  $form['buttons']['pay'] = array(
    '#type' => 'submit',
    '#value' => t('Pay via my bank'),
  );

  return $form;
}

/**
 * Validate handler for ideal_form_transaction().
 */
function ideal_form_transaction_validate(array $form, array &$form_state) {
  if (in_array($form_state['values']['issuer'], array('ideal_choose', 'ideal_choose_other'))) {
    form_set_error('issuer', t('Choose your bank.'));
  }
}

/**
 * Form submit handler for ideal_form_transaction().
 */
function ideal_form_transaction_submit(array $form, array &$form_state) {
  // @todo The original submit handler used _ideal_session_set(). Find out how
  // to reimplement it.
  $values = $form_state['values'];
  $transaction = $values['transaction'];
  $transaction->amount = $values['amount'];
  $transaction->pay($values['issuer']);
}

function theme_ideal_form_transaction(array $form) {
  drupal_add_js(drupal_get_path('module', 'ideal') . '/js/ideal.js');

  $test = isset($form['test']) ? '<div class="messages warning">' . $form['test']['#value'] . '</div>' : NULL;
  $form['test']['#printed'] = TRUE;
  $buttons = drupal_render($form['buttons']);
  $return = '<p class="description">' . $form['return']['#value'] . '</p>';
  $form['return']['#printed'] = TRUE;

  return drupal_render($form) . $test . $buttons . $return;
}

/**
 * Process a user who just returned from his issuer's website after a transaction.
 */
function ideal_transaction_return() {
  $entrance_code = $_GET['ec'];
  $transaction_id = $_GET['trxid'];
  $transaction = ideal_transaction_load(array(
    'transaction_id' => $transaction_id,
    'entrance_code' => $entrance_code,
    'status' => IDEAL_STATUS_OPEN,
  ));

  if (!$transaction) {
    drupal_access_denied();
    exit;
  }

  $transaction->receiveConsumer();
}

/**
 * An iDEAL transaction as a Rules data type.
 */
module_load_include('inc', 'rules', 'rules.data_types');
class iDEALRulesTransaction extends rules_data_type {
  function get_identifier() {
    $transaction = $this->get();

    return $transaction->itid;
  }
}